@@ -4,7 +4,7 @@ use warnings;
 
 package Path::Iterator::Rule;
 # ABSTRACT: Iterative, recursive file finder
-our $VERSION = '0.001'; # VERSION
+our $VERSION = '0.002'; # VERSION
 
 # Register warnings category
 use warnings::register;
@@ -16,7 +16,6 @@ use Data::Clone qw/data_clone/;
 use File::Basename qw/basename/;
 use List::Util qw/first/;
 use Number::Compare 0.02;
-use IO::Dir;
 use Scalar::Util qw/blessed/;
 use Text::Glob qw/glob_to_regex/;
 use Try::Tiny;
@@ -29,7 +28,7 @@ use namespace::clean;
 
 sub new {
     my $class = shift;
-    return bless { rules => [ sub { 1 } ] }, ref $class || $class;
+    return bless { rules => [ sub () { 1 } ] }, ref $class || $class;
 }
 
 sub clone {
@@ -56,7 +55,7 @@ sub add_helper {
         }
     }
     else {
-        Carp::carp( "Can't add rule '$name' because it conflicts with an existing method" );
+        Carp::carp("Can't add rule '$name' because it conflicts with an existing method");
     }
 }
 
@@ -67,15 +66,14 @@ sub add_helper {
 
 sub _objectify {
     my ( $self, $path ) = @_;
-    return $path;
+    return "$path";
 }
 
 sub _children {
     my $self = shift;
     my $path = "" . shift; # stringify objects
-    return () unless -d $path;
-    my $dir = IO::Dir->new($path);
-    return map { "$path/$_" } grep { $_ ne "." && $_ ne ".." } $dir->read;
+    opendir( my $dh, $path );
+    return map { [ $_, "$path/$_" ] } grep { $_ ne "." && $_ ne ".." } readdir $dh;
 }
 
 #--------------------------------------------------------------------------#
@@ -85,6 +83,7 @@ sub _children {
 my %defaults = (
     depthfirst      => 0,
     follow_symlinks => 1,
+    sorted          => 1,
     loop_safe       => ( $^O eq 'MSWin32' ? 0 : 1 ),       # No inode #'s on Windows
     error_handler   => sub { die sprintf( "%s: %s", @_ ) },
 );
@@ -96,7 +95,9 @@ sub iter {
       : ref( $_[-1] ) && !blessed( $_[-1] ) ? pop
       :                                       {};
     my $opts = { %defaults, %$args };
-    my @queue = map { { path => $self->_objectify($_), depth => 0 } } @_ ? @_ : '.';
+    my @queue =
+      map { { base => basename("$_"), path => $self->_objectify($_), depth => 0 } }
+      @_ ? @_ : '.';
     my $stash = {};
     my %seen;
 
@@ -104,7 +105,7 @@ sub iter {
         LOOP: {
             my $task = shift @queue
               or return;
-            my ( $item, $depth ) = @{$task}{qw/path depth/};
+            my ( $base, $item, $depth ) = @{$task}{qw/base path depth/};
             return $item->[0] if ref $item eq 'ARRAY'; # deferred for postorder
             my $string_item = "$item";
             if ( !$opts->{follow_symlinks} ) {
@@ -112,28 +113,36 @@ sub iter {
             }
             local $_ = $item;
             $stash->{_depth} = $depth;
-            my $interest =
-              try { $self->test( $item, $stash ) }
-            catch { $opts->{error_handler}->( $item, $_ ) };
+            my $interest;
+            if ( $opts->{error_handler} ) {
+                $interest =
+                  try { $self->test( $item, $base, $stash ) }
+                catch { $opts->{error_handler}->( $item, $_ ) };
+            }
+            else {
+                $interest = $self->test( $item, $base, $stash );
+            }
             my $prune = $interest && !( 0 + $interest ); # capture "0 but true"
             $interest += 0;                              # then ignore "but true"
-            my $unique_id = $self->_unique_id( $string_item, $opts );
 
-            if ( -d $string_item && !$seen{$unique_id}++ && !$prune ) {
+            if (   -d $string_item
+                && !$prune
+                && ( !$opts->{loop_safe} || $self->_is_unique( $string_item, $stash ) ) )
+            {
                 if ( !-r $string_item ) {
                     warnings::warnif("Directory '$string_item' is not readable. Skipping it");
                 }
                 elsif ( $opts->{depthfirst} ) {
-                    my @next = $self->_taskify( $depth + 1, $self->_children($item) );
+                    my @next = $self->_taskify( $opts, $depth + 1, $self->_children($item) );
                     # for postorder, requeue as reference to signal it can be returned
                     # without being retested
-                    push @next, { path => [$item], depth => $depth }
+                    push @next, { base => $base, path => [$item], depth => $depth }
                       if $interest && $opts->{depthfirst} > 0;
                     unshift @queue, @next;
                     redo LOOP if $opts->{depthfirst} > 0;
                 }
                 else {
-                    push @queue, $self->_taskify( $depth + 1, $self->_children($item) );
+                    push @queue, $self->_taskify( $opts, $depth + 1, $self->_children($item) );
                 }
             }
             return $item
@@ -167,10 +176,9 @@ sub or {
     my $self    = shift;
     my @rules   = $self->_rulify( "or", @_ );
     my $coderef = sub {
-        my $item = shift;
         my $result;
         for my $rule (@rules) {
-            $result = $rule->($item) || 0;
+            $result = $rule->(@_) || 0;
             return $result if $result; # want to shortcut on "0 but true"
         }
         return $result;
@@ -183,8 +191,7 @@ sub not {
     my @rules   = $self->_rulify( "not", @_ );
     my $obj     = $self->new->and(@rules);
     my $coderef = sub {
-        my $item   = shift;
-        my $result = $obj->test($item);
+        my $result = $obj->test(@_);
         # XXX what to do about "0 but true"? Ignore it?
         return $result ? "0" : "1";
     };
@@ -196,18 +203,17 @@ sub skip {
     my @rules   = $self->_rulify( "not", @_ );
     my $obj     = $self->new->or(@rules);
     my $coderef = sub {
-        my $item   = shift;
-        my $result = $obj->test($item);
+        my $result = $obj->test(@_);
         return $result ? "0 but true" : "1";
     };
     return $self->and($coderef);
 }
 
 sub test {
-    my ( $self, $item, $stash ) = @_;
+    my ( $self, $item, $base, $stash ) = @_;
     my $result;
     for my $rule ( @{ $self->{rules} } ) {
-        $result = $rule->( $item, $stash ) || 0;
+        $result = $rule->( $item, $base, $stash ) || 0;
         return $result if !( 0 + $result ); # want to shortcut on "0 but true"
     }
     return $result;
@@ -237,29 +243,27 @@ sub _rulify {
 }
 
 sub _taskify {
-    my ( $self, $depth, @paths ) = @_;
-    return map { { path => $_, depth => $depth } } sort { "$a" cmp "$b" } @paths;
+    my ( $self, $opts, $depth, @paths ) = @_;
+    if ( $opts->{sorted} ) {
+        @paths = sort { "$a->[0]" cmp "$b->[0]" } @paths;
+    }
+    return map { { base => $_->[0], path => $_->[1], depth => $depth } } @paths;
 }
 
-sub _unique_id {
-    my ( $self, $string_item, $opts ) = @_;
+sub _is_unique {
+    my ( $self, $string_item, $stash ) = @_;
     my $unique_id;
-    if ( $opts->{loop_safe} ) {
-        my @st = eval { stat $string_item };
-        @st = eval { lstat $string_item } unless @st;
-        if (@st) {
-            $unique_id = join( ",", $st[0], $st[1] );
-        }
-        else {
-            my $type = -d $string_item ? 'directory' : 'file';
-            warnings::warnif("Could not stat $type '$string_item'");
-            $unique_id = $string_item;
-        }
+    my @st = eval { stat $string_item };
+    @st = eval { lstat $string_item } unless @st;
+    if (@st) {
+        $unique_id = join( ",", $st[0], $st[1] );
     }
     else {
+        my $type = -d $string_item ? 'directory' : 'file';
+        warnings::warnif("Could not stat $type '$string_item'");
         $unique_id = $string_item;
     }
-    return $unique_id;
+    return !$stash->{_seen}{$unique_id}++;
 }
 
 #--------------------------------------------------------------------------#
@@ -302,31 +306,38 @@ while ( my ( $k, $v ) = each %simple_helpers ) {
     __PACKAGE__->add_helper( $k, sub { return $v } );
 }
 
+sub _generate_name_matcher {
+    my (@patterns) = @_;
+    if ( @patterns > 1 ) {
+        return sub {
+            my $name = "$_[1]";
+            return ( first { $name =~ $_ } @patterns ) ? 1 : 0;
+          }
+    }
+    else {
+        my $pattern = $patterns[0];
+        return sub {
+            my $name = "$_[1]";
+            return $name =~ $pattern ? 1 : 0;
+          }
+    }
+}
+
 # "complex" helpers take arguments
 my %complex_helpers = (
     name => sub {
         Carp::croak("No patterns provided to 'name'") unless @_;
-        my @patterns = map { _regexify($_) } @_;
-        return sub {
-            my $f    = shift;
-            my $name = basename "$f";
-            return ( first { $name =~ $_ } @patterns ) ? 1 : 0;
-          }
+        _generate_name_matcher( map { _regexify($_) } @_ );
     },
     iname => sub {
         Carp::croak("No patterns provided to 'iname'") unless @_;
-        my @patterns = map { _regexify( $_, "i" ) } @_;
-        return sub {
-            my $f    = shift;
-            my $name = basename "$f";
-            return ( first { $name =~ m{$_}i } @patterns ) ? 1 : 0;
-          }
+        _generate_name_matcher( map { _regexify( $_, "i" ) } @_ );
     },
     min_depth => sub {
         Carp::croak("No depth argument given to 'min_depth'") unless @_;
         my $min_depth = 0 + shift; # if this warns, do here and not on every file
         return sub {
-            my ( $f, $stash ) = @_;
+            my ( $f, $b, $stash ) = @_;
             return $stash->{_depth} >= $min_depth;
           }
     },
@@ -334,7 +345,7 @@ my %complex_helpers = (
         Carp::croak("No depth argument given to 'max_depth'") unless @_;
         my $max_depth = 0 + shift; # if this warns, do here and not on every file
         return sub {
-            my ( $f, $stash ) = @_;
+            my ( $f, $b, $stash ) = @_;
             return $stash->{_depth} <= $max_depth ? 1 : "0 but true"; # prune
           }
     },
@@ -362,8 +373,7 @@ __PACKAGE__->add_helper(
         Carp::croak("No patterns provided to 'skip_dirs'") unless @_;
         my $name_check = Path::Iterator::Rule->new->name(@_);
         return sub {
-            my $f = shift;
-            return "0 but true" if -d "$f" && $name_check->test($f);
+            return "0 but true" if -d "$_[0]" && $name_check->test(@_);
             return 1; # otherwise, like a null rule
           }
       } => 1 # don't create not_skip_dirs
@@ -496,7 +506,7 @@ Path::Iterator::Rule - Iterative, recursive file finder
 
 =head1 VERSION
 
-version 0.001
+version 0.002
 
 =head1 SYNOPSIS
 
@@ -552,7 +562,7 @@ follows symlinks (by default, but can be disabled)
 
 =item *
 
-directories visited only once (no infinite loops)
+directories visited only once (no infinite loop; can be disabled)
 
 =item *
 
@@ -610,7 +620,7 @@ C<depthfirst> -- Controls order of results.  Valid values are "1" (post-order, d
 
 =item *
 
-C<error_handler> -- Catches errors during execution of rule tests. Default handler dies with the filename and error.
+C<error_handler> -- Catches errors during execution of rule tests. Default handler dies with the filename and error. If set to undef, error handling is disabled.
 
 =item *
 
@@ -620,6 +630,10 @@ C<follow_symlinks> -- Follow directory symlinks when true. Default is 1.
 
 C<loop_safe> -- Prevents visiting the same directory more than once when true.  Default is 1.
 
+=item *
+
+C<sorted> -- Whether entries in a directory are sorted before processing. Default is 1.
+
 =back
 
 Filesystem loops might exist from either hard or soft links.  The C<loop_safe>
@@ -891,13 +905,18 @@ version.
 
 Rules are implemented as (usually anonymous) subroutines callbacks that return
 a value indicating whether or not the rule matches.  These callbacks are called
-with two arguments.  The first argument is a path, which is
+with three arguments.  The first argument is a path, which is
 also locally aliased as the C<$_> global variable for convenience in simple
 tests.
 
   $rule->and( sub { -r -w -x $_ } ); # tests $_
 
-The second argument is a hash reference that can be used to maintain state.
+The second argument is the basename of the path, which is useful for certain
+types of name checks:
+
+  $rule->and( sub { $_[1] =~ /foo|bar/ } ); "foo" or "bar" in basename;
+
+The third argument is a hash reference that can be used to maintain state.
 Keys beginning with an underscore are B<reserved> for C<Path::Iterator::Rule>
 to provide additional data about the search in progress.
 For example, the C<_depth> key is used to support minimum and maximum
@@ -932,7 +951,7 @@ a depth of 3:
 
   $rule->and(
     sub {
-      my ($path, $stash) = @_;
+      my ($path, $basename, $stash) = @_;
       return $stash->{_depth} <= 3 ? 1 : "0 but true";
     }
   );
@@ -959,15 +978,14 @@ if the filename is "foo":
   package Path::Iterator::Rule::Foo;
 
   use Path::Iterator::Rule;
-  use File::Basename qw/basename/;
 
   Path::Iterator::Rule->add_helper(
     foo => sub {
       my @args = @_; # do this to customize closure with arguments
       return sub {
-        my ($item) = shift;
+        my ($item, $basename) = shift;
         return if -d "$item";
-        return basename($item) =~ /^foo$/;
+        return $basename =~ /^foo$/;
       }
     }
   );
@@ -998,10 +1016,12 @@ _objectify -- given a path, return an object
 
 =item *
 
-_chilren -- given a directory, return an (unsorted) list of entries within it, excluding "." and ".."
+_children -- given a directory, return an (unsorted) list of [ basename, full path ] entries within it, excluding "." and ".."
 
 =back
 
+Note that C<_children> should return a I<list> of I<tuples>, where the tuples are basename and full path.
+
 See L<Path::Class::Rule> source for an example.
 
 =head1 LEXICAL WARNINGS
@@ -1013,6 +1033,22 @@ correct scope:
 
   no warnings 'Path::Iterator::Rule';
 
+=head1 PERFORMANCE
+
+By default, C<Path::Iterator::Rule> iterator options are "slow but safe".  They
+ensure uniqueness, return files in sorted order, and throw nice error messages
+if something goes wrong.
+
+If you want speed, set these options:
+
+    my %options = (
+        loop_safe => 0,
+        sorted => 0,
+        error_handler => undef
+    );
+
+    my $iter = $rule->iter( @dirs, \%options );
+
 =head1 CAVEATS
 
 Some features are still unimplemented:
